#!/usr/bin/env python
# coding:utf-8

"lrdgdal package"

__author__ = "Thomas Scharrenbach (thomas@scharrenbach.net)"
__copyright__ = "Copyright (C) 2014 Thomas Scharrenbach"
__license__ = "Apache License v2"
__version__ = "0.0.6"


# relies on nothing
from .gdalrdf import create_raster_from_file
from .gdalrdf import create_raster_from_iri
from .gdalrdf import create_raster_from_stream
from .gdalrdf import create_raster_from_data
from .gdalrdf import get_pixel_value

from . import create_raster_from_file
from . import create_raster_from_iri
from . import get_pixel_value

from .classes.BandRDFGraph import BandRDFGraph
from . import BandRDFGraph
from .classes.Raster import Raster
from . import Raster
from .classes.RasterRDFDataset import RasterRDFDataset
from . import RasterRDFDataset


__all__ = ['create_raster_from_file','create_raster_from_stream', 'create_raster_from_data',
           'create_raster_from_iri', 'get_pixel_value',
           'gdalutils', 'ogcutils', 'sparql',
           'Raster', 'BandRDFGraph', 'RasterRDFDataset']



#!/usr/bin/env python
# coding:utf-8
"""lrdgdal RasterGraph class"""

__author__ = "Thomas Scharrenbach (thomas@scharrenbach.net)"
__copyright__ = "Copyright (C) 2014 Thomas Scharrenbach"
__license__ = "Apache License v2"
__version__ = "0.0.6"

from rdflib import Graph, Dataset

from .. import gdalrdf

from .Raster import Raster


class RasterRDFDataset(Dataset):
    """An RDF dataset that holds multiple graphs storing meta-data and pixel-data."""

    def __init__(self, store='default'):
        """Stores RDF graphs along with raster data.

        :store see rdflib documentation

        Sets default_union to False, since it is not desirable to have pixel values accessable from the default graph.
        """
        super(RasterRDFDataset, self).__init__(store=store, default_union=False)

        self.__raster_map = dict()
        self.__band_map = dict()

        d = Dataset()
        d.add_graph(Graph())

        try:
            g = Graph()
            self.add_graph(g)
        except Exception as e:
            raise e

    def add_raster(self, raster):
        """Adds a lrdgdal.Raster.

        If the raster was already added to this dataset, an Exception will be raised.

        :param raster: the lrdgdal.Raster to add.
        :return: self
        """
        if not raster.iri in self.__raster_map:
            for t in gdalrdf.create_triples_for_raster(raster=raster):
                self.add(t)

            for band_iri, band in raster.bands.items():
                self.add_graph(band)
                for t in gdalrdf.create_triples_for_band(band=band, raster=raster):
                    self.add(t)
            self.__raster_map[raster.iri] = raster

        else:
            raise Exception('Raster <{}> already in raster graph!'.format(raster.iri))

        return self

    def parse(self, source=None, publicID=None, format='gdal', location=None, file=None, data=None, **args):
        """Read a raster or a regular RDF graph from a file, an IRI or an input stream and add it to the dataset.

        If the format does not match 'gdal', then parsing is delegated to the super class.

        The order in which the parameters are evaluated is:

        * source
        * file
        * location
        * data

        If all of the above are None, then an exception will be raised.

        If any of the above is not None, then the respective creation method will be tried. In case creation raises
        an Exception, this Exception will be forwarded.

        :param source: stream from which to load gdal dataset or RDF graph.
        :param publicID: IRI under which the loaded graph shall be available.
        :param format: use 'gdal' for gdal datasets; for other see the rdflib documentation.
        :param location: IRI from which to load gdal dataset or RDF graph.
        :param file: file from which to load gdal dataset or RDF graph.
        :param data: data from which to load gdal dataset or RDF graph.
        :param args: optional args for loading RDF graphs.
        :return: the parsed graph.
        """
        if format != 'gdal':
            super(RasterRDFDataset, self).parse(
                source=source, publicID=publicID, format=format, location=location, data=data, args=args)
        else:
            if publicID is None:
                raise Exception('Raster must have an IRI!')

            if source is not None:
                try:
                    self.add_raster(
                        Raster(dataset=gdalrdf.create_raster_from_stream(stream=source), raster_iri=publicID))
                except Exception as e:
                    raise e

            elif file is not None:
                try:
                    self.add_raster(Raster(dataset=gdalrdf.create_raster_from_file(filename=file), raster_iri=publicID))
                except Exception as e:
                    raise e

            elif location is not None:
                try:
                    self.add_raster(Raster(dataset=gdalrdf.create_raster_from_iri(iri=location), raster_iri=publicID))
                except Exception as e:
                    raise e

            elif data is not None:
                try:
                    self.add_raster(Raster(dataset=gdalrdf.create_raster_from_data(data=data), raster_iri=publicID))
                except Exception as e:
                    raise e

            else:
                raise Exception('Neither file, nor IRI nor input stream defined! Cannot create raster!')

    def get_raster_map(self):
        return self.__raster_map

    def get_band_map(self):
        return self.__band_map

    def get_raster(self, raster_iri):
        return self.__raster_map[raster_iri]

    def get_band(self, band_iri):
        return self.__band_map[band_iri]


#!/usr/bin/env python
# coding:utf-8

"""lrdgdal sparql module"""

from lrdgdal.classes import RasterRDFDataset

__author__ = "Thomas Scharrenbach (thomas@scharrenbach.net)"
__copyright__ = "Copyright (C) 2014 Thomas Scharrenbach"
__license__ = "Apache License v2"
__version__ = "0.0.6"

import logging

import binascii

from rdflib import BNode, Variable, Literal, URIRef

from rdflib.namespace import XSD

from lrdgdal.namespace import GDAL

from lrdgdal import BandRDFGraph, gdalrdf, gdalutils, ogcutils


def lrdgal_bgp_band_value(ctx, part):
    """Evaluate raster graph for pixel value of a band.

    :param ctx: input solution mappings and the graph
    :param part: the basic graph pattern
    :return: a set of solution mappings
    """
    if part.name == 'BGP':
        logging.debug('Evaluating BGP for "lrdgal_bgp_band_value"')
        if type(ctx.graph) is BandRDFGraph:
            logging.debug('Found graph of type "BandRDFGraph"')
            g = ctx.graph
            point = None
            spatial_ref_point = None

            non_gdal_triples = []
            band_iris = []

            # Iterator through the triple pattern and check whether it
            # contains all information necessary to perform a pixel lookup.
            for t in part.triples:
                if t[1] == GDAL.band_value:
                    band_value_var = t[2]
                    if type(t[0]) is Variable:
                        band_iris = [g.iri]
                    else:
                        band_iris.append(t[0])

                elif t[1] == GDAL.pixel_coordinates and type(t[2]) is Literal:
                    spatial_ref_point, point = ogcutils.parse_wktliteral(t[2])

                elif t[1] == GDAL.pixel_value and type(t[2]) is Variable:
                    pixel_value_var = t[2]

                else:
                    non_gdal_triples.append(t)

            if point is not None and spatial_ref_point is not None:
                c = None
                for band_iri in band_iris:
                    band = g
                    raster = band.raster
                    c = ctx.push()
                    c[band_value_var] = BNode()
                    c[pixel_value_var] = \
                        gdalrdf.get_pixel_value(
                            dataset=raster.dataset,
                            band_no=1,
                            spatial_ref_raster=raster.spatial_ref,
                            point=[point.GetX(), point.GetY()],
                            affine_trans_inv=raster.affine_trans_inv,
                            raster_dimensions=raster.raster_dimensions,
                            spatial_ref_point=spatial_ref_point
                        )
                # Delegate to normal BGP evaluation with remaining non_gdal
                # patterns.
                # NOTE: this import cannot be done outside this method!
                from rdflib.plugins.sparql.evaluate import evalBGP
                if c is not None:
                    return evalBGP(c, non_gdal_triples)
                else:
                    return ctx.solution()

            else:
                # Skip evaluation since the BGP does not hold all required
                # information.
                raise NotImplementedError()

        else:
            # Skip evaluation since the graph does not represent a raster band.
            raise NotImplementedError()

    else:
        # Skip evaluation since we only handle BGP matching.
        raise NotImplementedError()


def lrdgal_bgp_clip(ctx, part):
    """Clip a raster with a polygon geometry.

    :param ctx: input solution mappings and the graph
    :param part: the basic graph pattern
    :return: a set of solution mappings
    """
    if part.name == 'BGP':
        logging.debug('Evaluating BGP for "lrdgal_bgp_clip"')

        if type(ctx.graph) is BandRDFGraph:
            logging.debug('Found graph of type "BandRDFGraph"')
            band = ctx.graph
            value_mask_geometry = None
            non_gdal_triples = list()

            we_are_responsible = True
            var_band = None
            value_clip_value = None
            value_clip = None
            var_clip = None
            value_mask_geometry = None

            seen_clip_value = False
            seen_clip = False
            seen_mask_geometry = False

            # The BGP must have the following structure:
            #
            # (?VAR_X or IRI_X) lrdgdal:clipValue (?VAR_Y or IRI_Y or BNODE_Y) .
            # (?VAR_Y or IRI_Y or BNODE_Y) lrdgdal:clip (?VAR_Z or IRI_Z or BNODE_Z) .
            # (?VAR_Y or IRI_Y or BNODE_Y) lrdgdal:maskGeometry WKT_LITERAL .
            #
            for t in part.triples:
                if t[1] == GDAL.clip_value:
                    if seen_clip_value:
                        raise RuntimeError(
                            'Not more than one property with IRI "{}" allowed!'.format(GDAL.clip))
                    seen_clip_value = True
                    if type(t[0]) is Variable:
                        if ctx[t[0]] is not None:
                            if ctx[t[0]] != URIRef(band.iri):
                                we_are_responsible = False
                        else:
                            var_band = t[0]

                    # Check join condition
                    if value_clip_value is None:
                        # Save for checking join condition
                        value_clip_value = t[2]
                    else:
                        # if join condition not fulfilled, we are not responsible
                        if value_clip_value != t[2]:
                            we_are_responsible = False
                            logging.warning('No join for bgp possible, no results!')

                elif t[1] == GDAL.clip:
                    if seen_clip:
                        raise RuntimeError(
                            'Not more than one property with IRI "{}" allowed!'.format(GDAL.clip_value))
                    seen_clip = True

                    if type(t[2]) is Variable:
                        var_clip = t[2]
                    elif type(t[2]) is Literal:
                        value_clip = t[2]
                    else:
                        we_are_responsible = False
                        logging.warning('Clip must be variable or Literal, no results!')

                    # Check join condition
                    if value_clip_value is None:
                        # Save for checking join condition
                        value_clip_value = t[0]
                    else:
                        # if join condition not fulfilled, we are not responsible
                        if value_clip_value != t[0]:
                            we_are_responsible = False
                            logging.warning('No join for bgp possible, no results!')

                elif t[1] == GDAL.mask_geometry:
                    if seen_mask_geometry:
                        raise RuntimeError(
                            'Not more than one property with IRI "{}" allowed!'.format(GDAL.mask_geometry))
                    seen_mask_geometry = True

                    # Check join condition
                    if value_clip_value is None:
                        # Save for checking join condition
                        value_clip_value = t[0]
                    else:
                        # if join condition not fulfilled, we are not responsible
                        if value_clip_value != t[0]:
                            we_are_responsible = False
                            logging.warning('No join for bgp possible, no results!')

                    if type(t[2]) is Literal:
                        srs, value_mask_geometry = ogcutils.parse_wktliteral(t[2])
                        pass
                    else:
                        we_are_responsible = False
                        logging.warning('Mask geometry must be WKTLiteral, no results!')

                # Triples not handled by this BGP!
                else:
                    non_gdal_triples.append(t)

            if we_are_responsible and seen_clip_value and seen_clip and seen_mask_geometry:
                clip_data = gdalutils.clip_dataset(
                    dataset=band.raster.dataset , mask_geometry=value_mask_geometry,
                    bands=[band.band.GetBand()], nodata_value=-1, fill_value=0)
                bin_values = gdalrdf.geotiff2base64(dataset=clip_data)
                value_clip = Literal(binascii.hexlify(bin_values), datatype=XSD.hexBinary)
                if var_band is not None:
                    c = ctx.push()
                    c[var_band] = URIRef(band.iri)
                else:
                    c  = ctx
                if var_clip is not None:
                    c[var_clip] = value_clip
                if type(value_clip_value) is Variable:
                    c[value_clip_value] = BNode()
                from rdflib.plugins.sparql.evaluate import evalBGP
                return evalBGP(c, non_gdal_triples)

            else:
                logging.warning('No join for bgp possible, no results!')
                # Skip evaluation since the bgp is not eligible.
                raise NotImplementedError()

        else:
            # Skip evaluation since the graph does not represent a raster band.
            raise NotImplementedError()

    else:
        # Skip evaluation since we only handle BGP matching.
        raise NotImplementedError()


# Create a registry for the bgp extensions defines in this module.
_registry = dict()
_registry[GDAL.band_value] = lrdgal_bgp_band_value
_registry[GDAL.clip] = lrdgal_bgp_clip


def activate(modules='all'):
    """Activate the specified bgp extensions

    :param modules: IRIs of extensions to activate (default: all)
    :return:
    """
    from rdflib.plugins.sparql import CUSTOM_EVALS
    if modules == 'all':
        for key, value in _registry.items():
            CUSTOM_EVALS[key] = value
    else:
        for key in modules:
            if key in _registry:
                CUSTOM_EVALS[key] = _registry.get(key)



#!/usr/bin/env python
# coding:utf-8
"""lrdgdal utilities for gdal

This module contains mostly utilities for projecting points between different reference systems.
"""

__author__ = "Thomas Scharrenbach (thomas@scharrenbach.net)"
__copyright__ = "Copyright (C) 2014 Thomas Scharrenbach"
__license__ = "Apache License v2"
__version__ = "0.0.6"

import logging

import tempfile

import numpy

try:
    from osgeo import osr
    from osgeo import ogr
    from osgeo import gdal
except ImportError:
    import osr
    import ogr
    import gdal

WGS84 = osr.SpatialReference()
WGS84.SetWellKnownGeogCS("WGS84")


def project_coordinates(point, spatial_ref_target, spatial_ref_src=WGS84):
    """Transforms a geo-referenced point to real-valued target referenced system coordinates.

    :param point: the geo-reference point to transform
    :param spatial_ref_target: the spatial reference system of the target
    :param spatial_ref_src: the spatial reference system of the point (default WGS84)
    :return: point transformed to target coordinates
    """
    coordinate_trans = \
        osr.CoordinateTransformation(spatial_ref_src, spatial_ref_target)
    projected_point = \
        coordinate_trans.TransformPoint(point[0], point[1])

    # omit z-variable, since it has no meaning!
    return [projected_point[0], projected_point[1]]


def project_raster_to_geo(points, affine_trans):
    """Turns points in raster coordinates into a set of geo-referenced points.

    :param points: the points to transform
    :param affine_trans: the affine geo-transformation of the raster
    :return: array of points
    """
    x = numpy.array([[c[0] for c in points], [c[1] for c in points], [1 for c in points]])

    # TODO handle homogeneous coordinates zero values.
    return [[c[0]/c[2], c[1]/c[2]] for c in numpy.dot(affine_trans, x).T]


def project_geo_to_raster(points, affine_trans_inv):
    """Computes raster coordinates of a geo-referenced point.

    :param points: the points to project
    :param affine_trans_inv: the inverse of the raster's affine geo-transformation
    :return: True if point in raster, False else.

    Important: please make sure that the point is geo-referenced according to the
    raster's spatial reference system!
    """
    x = numpy.array([[c[0] for c in points], [c[1] for c in points], [1 for c in points]])

    # TODO handle homogeneous coordinates zero values.
    return [[c[0]/c[2], c[1]/c[2]] for c in numpy.dot(affine_trans_inv, x).T]


def point_in_raster(point, affine_trans_inv, raster_dimensions):
    """Checks whether a point (in raster coords) lies within the raster.

    :param point: the point to check
    :param affine_trans_inv: the inverse of the raster's affine geo-transformation
    :param raster_dimensions: [xsize, ysize, 1] as numpy array
    :return: True if point in raster, False else.

    Important: please make sure that you transformed the point into
    raster coordinates before applying this method!
    """
    r_c = project_geo_to_raster(
        points=[point],
        affine_trans_inv=affine_trans_inv)[0]
    in_raster = numpy.subtract(raster_dimensions, numpy.array([[r_c[0]], [r_c[1]], [1]]))

    # Return single truth value instead of numpy array.
    if round(in_raster[0][0]) >= 0 and round(in_raster[1][0]) >= 0:
        return True
    else:
        logging.warning('POINT NOT IN RASTER!')
        return False


def create_raster_base(
        driver="GTiff", raster_x_size=512, raster_y_size=256,
        geo_transform=[0, 1, 0, 0, 0, 1], srs=WGS84):
    """Base method for creating a raster.

    :param driver: driver of raster (default "GTiff")
    :param raster_x_size: x-size of raster (default 512)
    :param raster_y_size: y-size of raster (default 256)
    :param geo_transform: affine transformation (default [0,1,0,0,0,1])
    :param srs: spatial reference system (default WGS84)
    :return: raster without data.
    """
    driver = gdal.GetDriverByName(driver)
    if driver is None:
        raise Exception('Driver could not be instantiated: {}'.format(driver))

    dataset_file = tempfile.NamedTemporaryFile(delete=True)
    dataset = driver.Create(
        utf8_path=dataset_file.name, xsize=raster_x_size, ysize=raster_y_size,
        bands=1, eType=gdal.GDT_Byte)

    dataset.SetGeoTransform(geo_transform)
    dataset.SetProjection(srs.ExportToWkt())

    return dataset


def create_dataset_random(
        driver="MEM",
        raster_x_size=512,
        raster_y_size=512,
        geo_transform=[0, 1, 0, 0, 0, 1],
        srs=WGS84):
    """Create a gdal raster dataset of random numbers.

        :return: raster filled with random numbers
    """
    dataset = create_raster_base(
        driver=driver, raster_x_size=raster_x_size, raster_y_size=raster_y_size,
        geo_transform=geo_transform, srs=srs)

    raster = numpy.random.random_integers(0, 100, (dataset.RasterXSize, dataset.RasterYSize))
    dataset.GetRasterBand(1).WriteArray(raster)

    return dataset


def create_dataset_zeros(
        driver="GTiff", raster_x_size=512, raster_y_size=512, geo_transform=[0, 1, 0, 0, 0, 1], srs=WGS84):
    """Create a gdal raster dataset of zeros.

    :return: raster filled with zeros
    """
    dataset = create_raster_base(
        driver=driver, raster_x_size=raster_x_size, raster_y_size=raster_y_size,
        geo_transform=geo_transform, srs=srs)

    raster = numpy.zeros((dataset.RasterYSize, dataset.RasterXSize), dtype=numpy.uint8)
    dataset.GetRasterBand(1).WriteArray(raster)

    return dataset


def points2polygon(points):
    """Create a polygon from a list of points.

    :param points: list of 2D points
    :return: ogr polygon
    """
    # Create ring from list of points.
    ring = ogr.Geometry(ogr.wkbLinearRing)
    first = None
    for p in points:
        if first is None:
            first = p
        logging.debug('Adding point to geometry: {}'.format(p))
        ring.AddPoint_2D(p[0], p[1])
    ring.AddPoint_2D(first[0], first[1])

    # Create polygon
    poly = ogr.Geometry(ogr.wkbPolygon)
    poly.AddGeometry(ring)
    return poly


def affine_transformation(dataset):
    """Creates a numpy matrix holding the affine transformation in homogeneous coordinates.

    :param dataset: the gdal dataset
    :return: 3x3 numpy array with affine transformation
    """
    geo_trans = dataset.GetGeoTransform()
    return numpy.array([
        [geo_trans[1], geo_trans[2], geo_trans[0]],
        [geo_trans[4], geo_trans[5], geo_trans[3]],
        [0, 0, 1]
    ])


def affine_transformation_inv(dataset):
    """Creates a numpy matrix holding the inverse affine transformation in homogeneous coordinates.

    :param dataset: the gdal dataset
    :return: 3x3 numpy array with inverse affine transformation
    """
    return numpy.linalg.inv(affine_transformation(dataset=dataset))


def geometry_geo2raster(geometry, affine_trans_inv):
    """Projects a geometry into a raster.

    Note that the original object will be changed. In case this is not desired, please specify a clone as an argument
    to this method, for example, using Geometry.Clone().

    :param geometry: the geometry to transform
    :param affine_trans_inv: the inverse of the raster's affine transformation.
    :return: the modified geometry.
    """
    # Recursively process all sub-geometries
    for g in range(0, geometry.GetGeometryCount()):
        geometry_geo2raster(geometry.GetGeometryRef(g), affine_trans_inv)

    # Project each point into the raster's coordinate system.
    for p in range(0, geometry.GetPointCount()):
        point = geometry.GetPoint_2D(p)
        point = project_geo_to_raster([point], affine_trans_inv)[0]
        geometry.SetPoint_2D(point=p, x=point[0], y=point[1])

    return geometry


def bbox_for_geometry_inside_dataset(dataset, mask_geometry):
    """

    :param dataset:
    :param mask_geometry:
    :return:
    """

    # Create a copy of the mask mask_geometry and project it to the raster's SRS.
    geometry_raster = mask_geometry.Clone()
    geometry_raster.TransformTo(osr.SpatialReference(dataset.GetProjection()))

    # Project the mask geometry into raster coordinates.
    geometry_geo2raster(geometry_raster, affine_transformation_inv(dataset))

    # Convert the raster's bounding box to a mask_geometry in order to intersect it with the mask mask_geometry.
    bbox = points2polygon(
        [[0, 0],
        [0, dataset.RasterYSize],
        [dataset.RasterXSize, dataset.RasterYSize],
        [dataset.RasterXSize, 0]])

    # Get the bounding box of the intersection of the mask geometry with the raster's bounding box.
    # This will give us the bounding box for the clipping region, which we have to specifiy by
    # offset and size for gdal's methods for reading raster data.
    env = geometry_raster.Intersection(bbox).GetEnvelope()
    xoff = round(env[0])
    yoff = round(env[2])
    win_xsize = round(env[1])
    win_ysize = round(env[3])

    logging.debug('GEO_RASTER: {}'.format(geometry_raster))
    logging.debug('ENVELOPE_GEO_RASTER: {}'.format(env))

    # Due to rounding errors we have to check whether the computed values fir the raster's dimensions.
    # In case they do not fit we have to correct these.
    if xoff + win_xsize > dataset.RasterXSize:
        logging.debug('Correcting win_xsize from "{}" to "{}"'.format(
            win_xsize, dataset.RasterXSize - xoff))
        win_xsize = dataset.RasterXSize - xoff
    if yoff + win_ysize > dataset.RasterYSize:
        logging.debug('Correcting win_ysize from "{}" to "{}"'.format(
            win_ysize, dataset.RasterYSize - yoff))
        win_ysize = dataset.RasterYSize - yoff

    logging.debug('xSize: {}, ySize: {}'.format(dataset.RasterXSize, dataset.RasterYSize))
    logging.debug('xoff: {}, yoff:{}, win_x:{}, win_y:{}'.format(xoff, yoff, win_xsize, win_ysize))

    return xoff, yoff, win_xsize, win_ysize


def create_mask_dataset(dataset, mask_geometry, nodata_value=-1, gdal_driver_name='MEM', ogr_driver_name='Memory'):
    """Create a mask raster dataset from a mask geometry.

    :param dataset:
    :param mask_geometry:
    :param nodata_value:
    :param gdal_driver_name:
    :param ogr_driver_name:
    :return:
    """

    logging.info('Started creating mask...')
    # Compute dimensions of masked dataset.
    xoff, yoff, win_xsize, win_ysize = bbox_for_geometry_inside_dataset(
        dataset=dataset, mask_geometry=mask_geometry)

    # For creating the actual mask we will burn a shape layer into the raster.
    # To be more precise we will burn the shape into a new raster which has the dimensions of the
    # clipping region. We will use this auxiliary raster for creating the actual mask as a numpy array.
    #shape_layer = create_layer_for_shape(
    #    geometry=mask_geometry, srs=dataset.GetProjection(), ogr_driver_name=ogr_driver_name)

    ogr_driver = ogr.GetDriverByName(ogr_driver_name)

    if ogr_driver is None:
        raise Exception('Driver not available: "{}"'.format(ogr_driver_name))

    logging.info('Started creating layer for shape.')
    # Create an in-memory shape dataset holding the specified geometry.
    shape_layer_file = tempfile.NamedTemporaryFile(delete=True)
    shape_layer_file.close()
    ogr_ds = ogr_driver.CreateDataSource(shape_layer_file.name)
    shape_layer = ogr_ds.CreateLayer(
        name='point_out', srs=osr.SpatialReference(dataset.GetProjection()),
        geom_type=ogr.wkbMultiPolygon)
    feature_definition = shape_layer.GetLayerDefn()
    out_feature = ogr.Feature(feature_definition)
    out_feature.SetGeometry(mask_geometry)
    shape_layer.CreateFeature(out_feature)
    logging.info('Finished creating layer for shape.')

    gdal_driver = gdal.GetDriverByName(gdal_driver_name)
    if gdal_driver is None:
        raise Exception('Driver not available: "{}"'.format(gdal_driver_name))

    data_type = dataset.GetRasterBand(1).DataType

    # Create the masking dataset as a temporary file
    dst_file = tempfile.NamedTemporaryFile(delete=True)
    mask_ds = gdal_driver.Create(
        utf8_path=dst_file.name,
        xsize=win_xsize,
        ysize=win_ysize,
        bands=1,
        eType=data_type)
    mask_ds.SetProjection(dataset.GetProjection())

    # Translate affine transform according to offsets
    gt = dataset.GetGeoTransform()
    upper_left = project_raster_to_geo(
        [[xoff, yoff]], affine_trans=affine_transformation(dataset))[0]
    gt_new = (upper_left[0], gt[1], gt[2], upper_left[1], gt[4], gt[5])
    mask_ds.SetGeoTransform(gt_new)

    # Burn mask_geometry to raster layer.
    gdal.RasterizeLayer(dataset=mask_ds, bands=[1], layer=shape_layer, burn_values=[-1])
    dst_mask_band = mask_ds.GetRasterBand(1)
    dst_mask_band.SetNoDataValue(nodata_value)

    logging.info('Finished creating mask.')
    return mask_ds, xoff, yoff, win_xsize, win_ysize


def clip_bands(dataset, mask_geometry, bands='ALL', nodata_value=-1, fill_value=0,
               gdal_driver_name='MEM', ogr_driver_name='Memory'):
    """Clip the bands of a dataset with respect to a mask_geometry.

    :param dataset: raster
    :param mask_geometry: mask_geometry to be used for clip_dataset.
    :param bands: indices of bands to clip or 'ALL'
    :param nodata_value: value for no data for clip_dataset.
    :param gdal_driver_name: gdal driver to use for clipping and result (default MEM)
    :param ogr_driver_name: ogr driver to use for shape layer (default Memory)
    :return: list of numpy arrays, each of which holding the clipped data of the bands

    We perform the clipping by accomplishing the following steps:

    1. Project the geometry into raster coordinates.
    2. Compute the intersection between the projected geometry's bbox and the raster's bbox.
        This will serve as the clipping region.
    3. Create an ogr layer holding the mask geometry and burn it into a fresh raster, which has the dimensions of the
        clipping region.
    4. For each band of the raster, read the clipping region and mask it with the data from the fresh raster into which
        we burned the mask geometry.
    5. Write the result to a new raster with the same meta-data as the old raster. Note that this new raster holds only
        the original data of the clipping region. The affine transformation may have been corected by the translation
        imposed by the bounding box of the clipping region.
    """

    mask_ds, xoff, yoff, win_xsize, win_ysize = create_mask_dataset(
        dataset=dataset, mask_geometry=mask_geometry, nodata_value=nodata_value,
        gdal_driver_name=gdal_driver_name, ogr_driver_name=ogr_driver_name)

    dst_mask_band = mask_ds.GetRasterBand(1)
    mask_data = dst_mask_band.ReadAsArray(
        xoff=0, yoff=0, win_xsize=mask_ds.RasterXSize, win_ysize=mask_ds.RasterYSize)
    mask_ds = None

    band_indices = None
    if bands == 'ALL':
        band_indices = range(1, dataset.RasterCount+1)
    else:
        band_indices = bands

    result = dict()
    for band_idx in band_indices:
        # The indices of the bands holding the original data are identical to those used for storing intermediate
        # results.
        src_data_band = dataset.GetRasterBand(band_idx)

        src_data = src_data_band.ReadAsArray(
            xoff=xoff, yoff=yoff, win_xsize=win_xsize, win_ysize=win_ysize)

        result[band_idx] = numpy.ma.masked_array(
            src_data,
            numpy.logical_not(mask_data),
            fill_value=fill_value)

    return result, xoff, yoff, win_xsize, win_ysize


def clip_dataset(dataset, mask_geometry, bands='ALL', nodata_value=-1, fill_value=0, gdal_driver_name='MEM', ogr_driver_name='Memory'):
    """Clip dataset with respect to a mask_geometry.

    :param dataset: raster
    :param mask_geometry: mask_geometry to be used for clip_dataset.
    :param nodata_value: value for no data for clip_dataset.
    :param gdal_driver_name: gdal driver to use for clipping and result (default MEM)
    :param ogr_driver_name: ogr driver to use for shape layer (default Memory)
    :return: dataset
    """
    logging.info('Started creating result tiff.')

    gdal_driver = gdal.GetDriverByName(gdal_driver_name)
    if gdal_driver is None:
        raise Exception('Driver not available: "{}"'.format(gdal_driver_name))

    # Compute dimensions of masked dataset.
    xoff, yoff, win_xsize, win_ysize = bbox_for_geometry_inside_dataset(
        dataset=dataset, mask_geometry=mask_geometry)

    data_type = dataset.GetRasterBand(1).DataType

    # Create result dataset to temporary file.
    result_file = tempfile.NamedTemporaryFile(delete=True)
    logging.debug('Filename of result: {}'.format(result_file.name))
    result_ds = gdal_driver.Create(
        utf8_path=result_file.name,
        xsize=win_xsize,
        ysize=win_ysize,
        bands=dataset.RasterCount,
        eType=data_type)
    result_ds.SetProjection(dataset.GetProjection())
    #result_ds.SetGeoTransform(gt_new)
    result_ds.SetMetadata(dataset.GetMetadata())
    logging.info('Finished creating result tiff.')

    band_indices = range(1, dataset.RasterCount+1)
    masked_band_data, xoff, yoff, win_xsize, win_ysize = clip_bands(
        dataset=dataset, mask_geometry=mask_geometry, bands=band_indices,
        nodata_value=nodata_value, fill_value=fill_value,
        gdal_driver_name=gdal_driver_name, ogr_driver_name=ogr_driver_name)

    band_indices = None
    if bands == 'ALL':
        band_indices = range(1, dataset.RasterCount+1)
    else:
        band_indices = bands

    dst_band_idx = 0
    # Mask all bands of the original dataset
    for src_band_idx in band_indices:
        dst_band_idx += 1
        # The indices of the bands holding the original data are identical to those used for storing intermediate
        # results.
        src_data_band = dataset.GetRasterBand(src_band_idx)

        # Fill bands of result dataset with masked values.
        result_band = result_ds.GetRasterBand(dst_band_idx)
        # TODO check which meta-data has to be copied and which not
        #result_band.SetColorTable(src_data_band.GetColorTable())
        #result_band.SetRasterColorTable(src_data_band.GetRasterColorTable())
        result_band.SetRasterColorInterpretation(src_data_band.GetRasterColorInterpretation())
        result_band.SetColorInterpretation(src_data_band.GetColorInterpretation())
        #result_band.SetDefaultRAT(src_data_band.GetDefaultRAT())
        #result_band.SetCategoryNames(src_data_band.GetCategoryNames())
        result_band.SetNoDataValue(nodata_value)
        logging.info('Started masking data for band {}'.format(src_band_idx))
        result_band.WriteArray(masked_band_data.get(src_band_idx).filled(fill_value=nodata_value))
        logging.info('Finished masking data for band {}'.format(src_band_idx))

    return result_ds



#!/usr/bin/env python
# coding:utf-8

"""lrdgdal gdalrdf module"""



__author__ = "Thomas Scharrenbach (thomas@scharrenbach.net)"
__copyright__ = "Copyright (C) 2014 Thomas Scharrenbach"
__license__ = "Apache License v2"
__version__ = "0.0.6"

import tempfile

from rdflib import Graph, Literal, URIRef
from rdflib.namespace import RDF, XSD

import lxml.etree as etree

try:
    from urllib.request import urlretrieve
except:
    from urllib import urlretrieve

try:
    from osgeo import gdal
    from osgeo import osr
except ImportError:
    try:
        import gdal
    except Exception as e:
        logging.error('Could not load gdal package!')
        raise e

import logging

from lrdgdal import gdalutils
from .namespace import GDAL

_gdal2xsd = {'Byte': XSD.byte}


def gdaldatatype_to_xsd(gdal_datatype):
    """Translates gdal type names to XSD types.

    :param gdal_datatype:
    :return:
    """
    if gdal_datatype in _gdal2xsd.keys():
        return _gdal2xsd.get(gdal_datatype)
    else:
        return None


def create_raster_from_file(filename):
    """Open a gdal dataset from a file.

    :param filename: name of the raster file to open
    :return: gdal dataset
    """
    dataset = gdal.Open(filename)
    return dataset


def create_raster_from_iri(iri):
    """Open a gdal dataset from an IRI.

    :param iri: IRI to use for downloading the raster file
    :return: gdal dataset
    """
    tmp_file = tempfile.NamedTemporaryFile(delete=True)
    urlretrieve(iri, tmp_file.name)
    return create_raster_from_file(filename=tmp_file.name)


def create_raster_from_stream(stream):
    """Open a gdal dataset from input stream stream.

    :param stream: stream stream
    :return: gdal dataset

    The stream is first written to a tempfile that is deleted on exit. Then we try to open this tempfile using the gdal
    library.
    """
    tmp_file = tempfile.NamedTemporaryFile(delete=True)
    tmp_file.write(stream.read())
    return create_raster_from_file(filename=tmp_file.name)


def create_raster_from_data(data):
    """Open a gdal dataset from input data.

    :param data: input data
    :return: gdal dataset

    The data is first written to a tempfile that is deleted on exit. Then we try to open this tempfile using the gdal
    library.
    """
    tmp_file = tempfile.NamedTemporaryFile(delete=True)
    tmp_file.write(data=data)
    return create_raster_from_file(filename=tmp_file.name)


def create_triples_for_raster(raster):
    """Generate RDF triples for a raster.

    :param raster: lrdgdal.Raster
    :return: list of RDF triples
    """
    g_raster_iri = raster.iri
    g_dataset = raster.dataset
    triples = list()
    triples.append((g_raster_iri, RDF.type, GDAL.raster))

    # Add dimensions of raster
    triples.append((g_raster_iri, GDAL.rasterXSize, Literal(g_dataset.RasterXSize, datatype=XSD.integer)))
    triples.append((g_raster_iri, GDAL.rasterYSize, Literal(g_dataset.RasterYSize, datatype=XSD.integer)))

    # Add description string of raster
    triples.append((g_raster_iri, GDAL.description, Literal(g_dataset.GetDescription(), datatype=XSD.string)))

    _describe_spatial_ref(raster)

    # add gcps if any
    # TODO

    return triples


def create_band_iri(raster_iri, band_no):
    """Generic method for generating the correct IRI for a raster's band.

    :param raster_iri: iri of raster.
    :param band_no: number of the band.
    :return: IRI relative to the raster's IRI.
    """
    return URIRef('{}band{}/'.format(raster_iri, band_no))


def create_triples_for_band(band, raster):
    """Generate RDF triples for a band.

    :param band: lrdgdal.Band
    :param raster: lrdgdal.Raster
    :return: list of RDF triples
    """
    band_iri = band.iri
    raster_iri = raster.iri
    src_band = band.band

    triples = list()
    triples.append((band_iri, RDF.type, GDAL.band))
    triples.append((band_iri, GDAL.band, raster_iri))

    bs = src_band.GetBlockSize()
    triples.append((band_iri, GDAL.blockXSize, Literal(bs[0], datatype=XSD.integer)))
    triples.append((band_iri, GDAL.blockYSize, Literal(bs[1], datatype=XSD.integer)))

    category_names = src_band.GetCategoryNames()
    if category_names is not None:
        pass

    colour_interpretation = src_band.GetColorInterpretation()
    if colour_interpretation is not None:
        triples.append((band_iri, GDAL.color_interpretation, Literal(GDAL.create_gci(colour_interpretation))))

    triples.append((band_iri, GDAL.color_interpretation, Literal(GDAL.create_gdt(src_band.DataType))))

    mask_band_no = src_band.GetMaskBand().GetBand()
    if mask_band_no is not None:
        triples.append((band_iri, GDAL.mask_band, create_band_iri(raster_iri=raster.iri, band_no=mask_band_no)))

    # rat = src_band.GetDefaultRAT()
    # rat = gdal.RasterAttributeTable()

    mask_flags = src_band.GetMaskFlags()
    if mask_flags is not None:
        triples.append((band_iri, GDAL.mask_flags, Literal(mask_flags, datatype=XSD.integer)))
    src_band.GetHistogram(
        min=-0.5,
        max=255.5,
        buckets=256,
        include_out_of_range=0,
        approx_ok=1,
        callback=None,
        callback_data=None)
    src_band.HasArbitraryOverviews()
    src_band.GetCategoryNames()

    nodata_value = src_band.GetNoDataValue()
    if nodata_value is not None:
        triples.append((band_iri, GDAL.bandNoDataValue, Literal(nodata_value)))

    minimum = src_band.GetMinimum()
    if minimum is not None:
        triples.append((band_iri, GDAL.bandMin, Literal(minimum)))

    maximum = src_band.GetMaximum()
    if maximum is not None:
        triples.append((band_iri, GDAL.bandMax, Literal(maximum)))

    scale = src_band.GetScale()
    if scale is not None:
        triples.append((band_iri, GDAL.bandScale, Literal(scale)))

    unit_type = src_band.GetUnitType()
    if unit_type is not None and unit_type != '':
        triples.append((band_iri, GDAL.band_unit_type, Literal(unit_type)))

    ctable = src_band.GetColorTable()
    if ctable is None:
        logging.info('No ColorTable found')

    else:
        logging.debug("[ COLOR TABLE COUNT ] = {}".format(ctable.GetCount()))
        for i in range(0, ctable.GetCount()):
            entry = ctable.GetColorEntry(i)
            if not entry:
                continue
            logging.debug("[ COLOR ENTRY RGB ] = {}".format(ctable.GetColorEntryAsRGB(i, entry)))

    stats = src_band.GetStatistics(True, True)
    if stats is not None:
        triples.append((band_iri, GDAL.bandStatsMin, Literal(stats[0])))
        triples.append((band_iri, GDAL.bandStatsMax, Literal(stats[1])))
        triples.append((band_iri, GDAL.bandStatsMean, Literal(stats[2])))
        triples.append((band_iri, GDAL.bandStatsStdDev, Literal(stats[3])))
    return triples


def get_pixel_value(
        dataset,
        band_no,
        spatial_ref_raster,
        point,
        affine_trans_inv,
        raster_dimensions,
        spatial_ref_point=gdalutils.WGS84):
    """get_pixel_value(dataset, band_no, spatial_ref_raster, point, affine_trans_inv, raster_dimensions,
    spatial_ref_point=gdalutils.WGS84)

    Get pixel value of a point for a band of a raster.

    :param dataset: gdal raster dataset
    :param band_no: number of the dataset's band to query
    :param spatial_ref_raster: the srs of the raster dataset
    :param point: point as array
    :param affine_trans_inv: inverse matrix of affine transformation
    :param raster_dimensions: [x, y]
    :param spatial_ref_point: the srs of the point (default WGS84)
    :return: the pixel value as rdf literal
    """
    band = dataset.GetRasterBand(band_no)
    point_geo_coordinates = gdalutils.project_coordinates(
        point=point,
        spatial_ref_target=spatial_ref_raster,
        spatial_ref_src=spatial_ref_point)

    point_raster_coordinates = gdalutils.project_geo_to_raster(
        points=[point_geo_coordinates],
        affine_trans_inv=affine_trans_inv)[0]

    if gdalutils.point_in_raster(
            point=point_geo_coordinates, affine_trans_inv=affine_trans_inv,
            raster_dimensions=raster_dimensions):
        # Let's not forget to subtract one to match zero-based indices.
        raster_pixel_value = \
            band.ReadAsArray(
                int(round(point_raster_coordinates[0]-1)),
                int(round(point_raster_coordinates[1]-1)), 1, 1)

        if raster_pixel_value is not None:
            xsd_type = \
                gdaldatatype_to_xsd(gdal.GetDataTypeName(band.DataType))
            if xsd_type is None:
                xsd_type = XSD.string

            return Literal(raster_pixel_value[0][0], datatype=xsd_type)
        else:
            return None
    else:
        logging.warning('POINT NOT IN RASTER! {}'.format(point_raster_coordinates))
        return None


def geotiff2base64(dataset):
    """Transforms a dataset to an xsd:base64binary literal.

    :param dataset: the dataset to transform to an xsd:base64binary rdf literal.
    :return: rdf literal of type base64binary
    """
    gdal_driver_name = 'GTIFF'
    gdal_driver = gdal.GetDriverByName(gdal_driver_name)
    if gdal_driver is None:
        raise Exception('Driver not available: "{}"'.format(gdal_driver_name))
    with tempfile.NamedTemporaryFile(delete=True) as result_file:
        result_final = gdal_driver.CreateCopy(
            utf8_path=result_file.name,
            src=dataset, strict=0, options=['COMPRESS=LZW'])
        result_final = None
        return result_file.read()


def _describe_spatial_ref(raster):
    triples = []
    dataset = raster.dataset
    raster_iri = raster.iri

    projection = dataset.GetProjection()
    spatial_ref = osr.SpatialReference(projection)

    dom = etree.fromstring(spatial_ref.ExportToXML().replace(
        ">", " xmlns:gml=\"http://www.opengis.net/gml/\""
             " xmlns:xlink=\"http://www.w3.org/1999/xlink\" "
             "xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\">", 1))


    logging.debug(etree.tostring(dom))

    from pkg_resources import resource_string
    xml2rdf = resource_string(__name__, 'xml2rdf.xsl')
    xslt = etree.fromstring(xml2rdf)

    transform = etree.XSLT(xslt)
    newdom = transform(dom)

    sr = Graph().parse(data=newdom)

    for s in sr.subjects(RDF.type, URIRef('http://www.opengis.net/gml/ProjectedCRS')):
        triples.append((raster_iri, GDAL.projection, s))
    return triples



#!/usr/bin/env python
# coding:utf-8

"""lrdgadl GEOSPARQL RDF constants"""

__author__ = "Thomas Scharrenbach (thomas@scharrenbach.net)"
__copyright__ = "Copyright (C) 2014 Thomas Scharrenbach"
__license__ = "Apache License v2"
__version__ = "0.0.6"

from rdflib import URIRef

NS = 'http://www.opengis.net/geosparql#'

PREFERRED_PREFIX = 'ogc'

WKT_LITERAL = URIRef(NS + 'wktLiteral')


#!/usr/bin/env python
# coding:utf-8

"""lrdgdal namespace package"""

__author__ = "Thomas Scharrenbach (thomas@scharrenbach.net)"
__copyright__ = "Copyright (C) 2014 Thomas Scharrenbach"
__license__ = "Apache License v2"
__version__ = "0.0.6"

from . import GDAL

from . import GEOSPARQL

__all__ = ['GDAL', 'GEOSPARQL']


#!/usr/bin/env python
# coding:utf-8

"""lrdgdal Raster class"""

__author__ = "Thomas Scharrenbach (thomas@scharrenbach.net)"
__copyright__ = "Copyright (C) 2014 Thomas Scharrenbach"
__license__ = "Apache License v2"
__version__ = "0.0.6"

from rdflib import URIRef

from osgeo import osr

from lrdgdal import gdalutils
from lrdgdal import gdalrdf

from lrdgdal import BandRDFGraph

import numpy

from numpy.linalg import inv


class Raster(object):
    """An object that holds various data for a gdal dataset.

    This class is essentially a wrapper for a gdal raster object with
    extended utility functions.
    """

    def __init__(self, dataset, raster_iri):
        """Initializes all values for meta-data and geo-processing.

        :param dataset: a gdal dataset.

        Stores the affine geo-transformation and its inverse for later
        processing.
        """

        self.__dataset = dataset

        self.__iri = URIRef(raster_iri)

        self.__geo_trans = self.dataset.GetGeoTransform()

        self.__affine_trans = numpy.array([
            [self.__geo_trans[1], self.__geo_trans[2], self.__geo_trans[0]],
            [self.__geo_trans[4], self.__geo_trans[5], self.__geo_trans[3]],
            [0, 0, 1]
        ])

        self.__affine_trans_inv = inv(self.__affine_trans)

        self.__spatial_ref = osr.SpatialReference(self.dataset.GetProjection())

        self.__raster_dimensions = numpy.array([
            [self.dataset.RasterXSize],
            [self.dataset.RasterYSize],
            [1]
        ])

        self.__bands = dict()

        # TODO
        #geo_trans = self.dataset.GetGeoTransform()

        for band_no in range(self.dataset.RasterCount):
            band_no += 1

            try:
                src_band = self.dataset.GetRasterBand(band_no)
            except RuntimeError:
                raise Exception('No band {} found'.format(band_no))

            self.add_band(BandRDFGraph(
                band=src_band,
                raster=self,
                iri=gdalrdf.create_band_iri(raster_iri=self.iri, band_no=band_no)))

    def __exit__(self, type, value, traceback):
        """Cleans up, in particular deletes the reference to the raster dataset.

        :param type:
        :param value:
        :param traceback:
        """
        self.__dataset = None

    def add_band(self, b):
        self.__bands[b.iri] = b

    def geopoint_to_raster_coordinates(self, point, spatial_ref_point=gdalutils.WGS84):
        """geopoint_to_raster_coordinates(point, spatial_ref_point=gdalutils.WGS84)

        Transforms a geo-referenced point to real-valued raster coordinates.

        :param point: the geo-reference point to transform
        :param spatial_ref_point: the spatial reference system of the point (default WGS84)
        :return: point transformed to raster coordinates
        """
        point_geo_coordinates = gdalutils.project_coordinates(
            point=point,
            spatial_ref_target=self.__spatial_ref,
            spatial_ref_src=spatial_ref_point)
        return gdalutils.project_geo_to_raster(
            [point_geo_coordinates],
            self.__affine_trans_inv)

    def rasterbb_to_geo(self, spatial_ref_target=None):
        """Turns the raster's bounding box into a set of geo-referenced points.

        :param spatial_ref_target: the spatial reference system of the point (default None)
        :return: a list of points (2D arrays) representing the raster's bounding box.
        """
        result = gdalutils.project_raster_to_geo(
            points=self.bbox,
            affine_trans=self.__affine_trans)

        if spatial_ref_target is not None:
            result = [
                gdalutils.project_coordinates(
                    point=c,
                    spatial_ref_src=self.__spatial_ref,
                    spatial_ref_target=spatial_ref_target)
                for c in result]

        return result

    def point_in_raster(self, point):
        """Checks whether a point (in raster coords) lies within the raster.

        Important: please make sure that you transformed the point into
        raster coordinates before applying this method!

        :param point: the point to check
        :return: True if point in raster, False else.
        """
        import logging
        logging.warning('Raster point in raster...')
        return gdalutils.point_in_raster(
            point, self.__affine_trans_inv, self.__raster_dimensions)

    @property
    def dataset(self):
        """The raster's gdal raster dataset."""
        return self.__dataset

    @property
    def spatial_ref(self):
        """The raster's gdal spatial reference."""
        return self.__spatial_ref

    @property
    def affine_trans(self):
        """The raster's gdal affine transformation matrix."""
        return self.__affine_trans

    @property
    def affine_trans_inv(self):
        """The raster's gdal inverse affine transformation matrix."""
        return self.__affine_trans_inv

    @property
    def raster_dimensions(self):
        """The raster's dimensions [x, y]."""
        return self.__raster_dimensions

    @property
    def iri(self):
        """The raster's IRI under which it is available in an lrdgdal.RasterGraph."""
        return self.__iri

    @property
    def bands(self):
        """Map of pairs <band_iri, band> of the raster's bands."""
        return self.__bands

    @property
    def bbox(self):
        """
        :return: bounding box as a list of points in raster coordinates
        """
        return [
            [1, 1],
            [1, self.dataset.RasterYSize],
            [self.dataset.RasterXSize, self.dataset.RasterYSize],
            [self.dataset.RasterXSize, 1]]


#!/usr/bin/env python
# coding:utf-8
"""lrdgdal utilities for ogc and rdf

This module contains mostly utilities for transforming values between ogc and
rdf.
"""

__author__ = "Thomas Scharrenbach (thomas@scharrenbach.net)"
__copyright__ = "Copyright (C) 2014 Thomas Scharrenbach"
__license__ = "Apache License v2"
__version__ = "0.0.6"

import logging

import re

try:
    from osgeo import ogr, osr
except ImportError:
    try:
        import gdal
        import ogr
        import osr
    except Exception as e:
        logging.error('Could not load gdal package!')
        raise e

from lxml import etree

WGS84_IRI = 'http://www.opengis.net/def/crs/OGC/1.3/CRS84'
WGS84_URN = 'urn:ogc:def:crs:OGC:1.3:CRS84'
WKT_RE = r'(<.*>)[ ]+(.*)'


def parse_wktliteral(wkt_literal):
    """Parses a WKT RDF Literal value.

    WKT RDFS Literals have string values of "<IRI> WKT".
    IRI must identify a valid spatial reference system that GDAL can parse.

    :param wkt_literal: the Literal to parse.
    :return: pair: srs, ogr geometry
    """

    pattern = re.compile(WKT_RE)
    match = pattern.match(wkt_literal)
    if match is None:
        raise Exception('Not a valid WKT literal: {}'.format(wkt_literal))

    spatial_ref = osr.SpatialReference()
    spatial_ref.ImportFromWkt(osr.GetUserInputAsWKT(WGS84_IRI))

    return spatial_ref, ogr.CreateGeometryFromWkt(match.group(2))


def encode_as_wktliteral(spatial_ref, geometry):
    """Encode a geometry as an WKS RDF Literal.

    :param spatial_ref: the spatial reference system
    :param geometry:
    :return: the string value of the RDF Literal.
    """
    spatial_ref_iri = extract_iri_from_srs(spatial_ref)
    return '<{}> {}'.format(spatial_ref_iri, str(geometry))


def extract_iri_from_srs(spatial_ref):
    """Extract the IRI of a spatial reference system.

    :param spatial_ref: a spatial reference system
    :return: IRI of the spatial reference system
    """
    # Add proper namespace to GML representation of spatial reference system.
    tree = etree.fromstring(
        spatial_ref.ExportToXML().replace(
            ">",
            " xmlns:gml=\"http://www.opengis.net/gml/\""
            + " xmlns:xlink=\"http://www.w3.org/1999/xlink\""
            + " xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\">",
            1)
    )

    # Retrieve IRI via xpath query.
    result = tree.xpath(
        '//gml:srsID[parent::gml:GeographicCRS|parent::gml:ProjectedCRS]/gml:name',
        namespaces={'gml': 'http://www.opengis.net/gml/'})
    return '{}{}'.format(
        result[0].get('{http://www.opengis.net/gml/}codeSpace'),
        result[0].text)


#!/usr/bin/env python
# coding:utf-8

"""lrdgadl GDAL RDF constants"""

__author__ = 'Thomas Scharrenbach (thomas@scharrenbach.net)'
__copyright__ = 'Copyright (C) 2014 Thomas Scharrenbach'
__license__ = 'Apache License v2'
__version__ = '0.0.6'

try:
    from osgeo import gdal
except ImportError:
    import gdal


from rdflib import URIRef

#The base namespace for gdal rdf graphs.
NS = 'http://scharrenbach.net/gdal#'

#The base namespace for gdal drivers.
NS_DRIVERS = 'http://scharrenbach.net/gdal/drivers#'

#URI for band types.
band = URIRef(NS + 'Band')

rasterXSize = URIRef(NS + 'rasterXSize')

rasterYSize = URIRef(NS + 'rasterYSize')

blockXSize = URIRef(NS + 'blockXSize')

blockYSize = URIRef(NS + 'blockYSize')

bandNoDataValue = URIRef(NS + 'bandNoDataValue')

bandMin = URIRef(NS + 'bandMin')

bandMax = URIRef(NS + 'bandMax')

bandStatsMin = URIRef(NS + 'bandStatsMin')

bandStatsMax = URIRef(NS + 'bandStatsMax')

bandStatsMean = URIRef(NS + 'bandStatsMean')

bandStatsStdDev = URIRef(NS + 'bandStatsStdDev')

bandScale = URIRef(NS + 'bandScale')

band_unit_type = URIRef(NS + 'band_unit_type')

mask_band = URIRef(NS + 'mask_band')

mask_flags = URIRef(NS + 'mask_flags')

color_interpretation = URIRef(NS + 'color_interpretation')

dataType = URIRef(NS + 'dataType')

raster = URIRef(NS + 'Raster')

projection = URIRef(NS + 'projection')

driver = URIRef(NS + 'Driver')

geo_transform = URIRef(NS + 'geo_transform')

gcp = URIRef(NS + 'Gcp')

meta_dict = URIRef(NS + 'meta_dict')

meta_key = URIRef(NS + 'meta_key')

meta_value = URIRef(NS + 'meta_value')

short_name = URIRef(NS + 'short_name')

long_name = URIRef(NS + 'long_name')

description = URIRef(NS + 'description')

band_value = URIRef(NS + 'bandValue')

clip = URIRef(NS + 'clip')

clip_value = URIRef(NS + 'clipValue')

mask_geometry = URIRef(NS + 'maskGeometry')

pixel_value = URIRef(NS + 'pixelValue')

pixel_coordinates = URIRef(NS + 'pixelCoordinates')


GCI_AlphaBand = URIRef(NS + 'GCI_AlphaBand')
GCI_BlackBand = URIRef(NS + 'GCI_BlackBand')
GCI_BlueBand = URIRef(NS + 'GCI_BlueBand')
GCI_CyanBand = URIRef(NS + 'GCI_CyanBand')
GCI_GrayIndex = URIRef(NS + 'GCI_GrayIndex')
GCI_GreenBand = URIRef(NS + 'GCI_GreenBand')
GCI_HueBand = URIRef(NS + 'GCI_HueBand')
GCI_LightnessBand = URIRef(NS + 'GCI_LightnessBand')
GCI_MagentaBand = URIRef(NS + 'GCI_MagentaBand')
GCI_PaletteIndex = URIRef(NS + 'GCI_PaletteIndex')
GCI_RedBand = URIRef(NS + 'GCI_RedBand')
GCI_SaturationBand = URIRef(NS + 'GCI_SaturationBand')
GCI_Undefined = URIRef(NS + 'GCI_Undefined')
GCI_YCbCr_CbBand = URIRef(NS + 'GCI_YCbCr_CbBand')
GCI_YCbCr_CrBand = URIRef(NS + 'GCI_YCbCr_CrBand')
GCI_YCbCr_YBand = URIRef(NS + 'GCI_YCbCr_YBand')
GCI_YellowBand = URIRef(NS + 'GCI_YellowBand')


def create_gci(gci):
    if gci == gdal.GCI_AlphaBand:
        return GCI_AlphaBand
    elif gci == gdal.GCI_BlackBand:
        return GCI_BlackBand
    elif gci == gdal.GCI_BlueBand:
        return GCI_BlueBand
    elif gci == gdal.GCI_CyanBand:
        return GCI_CyanBand
    elif gci == gdal.GCI_GrayIndex:
        return GCI_GrayIndex
    elif gci == gdal.GCI_GreenBand:
        return GCI_GreenBand
    elif gci == gdal.GCI_HueBand:
        return GCI_HueBand
    elif gci == gdal.GCI_LightnessBand:
        return GCI_LightnessBand
    elif gci == gdal.GCI_MagentaBand:
        return GCI_MagentaBand
    elif gci == gdal.GCI_PaletteIndex:
        return GCI_PaletteIndex
    elif gci == gdal.GCI_RedBand:
        return GCI_RedBand
    elif gci == gdal.GCI_SaturationBand:
        return GCI_SaturationBand
    elif gci == gdal.GCI_Undefined:
        return GCI_Undefined
    elif gci == gdal.GCI_YCbCr_CbBand:
        return GCI_YCbCr_CbBand
    elif gci == gdal.GCI_YCbCr_CrBand:
        return GCI_YCbCr_CrBand
    elif gci == gdal.GCI_YCbCr_YBand:
        return GCI_YCbCr_YBand
    elif gci == gdal.GCI_YellowBand:
        return GCI_YellowBand
    else:
        return GCI_Undefined
        
GDT_Byte = URIRef(NS + 'GDT_Byte')
GDT_CFloat32 = URIRef(NS + 'GDT_CFloat32')
GDT_CFloat64 = URIRef(NS + 'GDT_CFloat64')
GDT_CInt16 = URIRef(NS + 'GDT_CInt16')
GDT_CInt32 = URIRef(NS + 'GDT_CInt32')
GDT_Float32 = URIRef(NS + 'GDT_Float32')
GDT_Float64 = URIRef(NS + 'GDT_Float64')
GDT_Int16 = URIRef(NS + 'GDT_Int16')
GDT_TypeCount = URIRef(NS + 'GDT_TypeCount')
GDT_UInt16 = URIRef(NS + 'GDT_UInt16')
GDT_UInt32 = URIRef(NS + 'GDT_UInt32')
GDT_Unknown = URIRef(NS + 'GDT_Unknown')


def create_gdt(gdt):
    if gdt == gdal.GDT_Byte:
        return GDT_Byte
    elif gdt == gdal.GDT_CFloat32:
        return GDT_CFloat32
    elif gdt == gdal.GDT_CFloat64:
        return GDT_CFloat64
    elif gdt == gdal.GDT_CInt16:
        return GDT_CInt16
    elif gdt == gdal.GDT_CInt32:
        return GDT_CInt32
    elif gdt == gdal.GDT_Float32:
        return GDT_Float32
    elif gdt == gdal.GDT_Float64:
        return GDT_Float64
    elif gdt == gdal.GDT_Int16:
        return GDT_Int16
    elif gdt == gdal.GDT_TypeCount:
        return GDT_TypeCount
    elif gdt == gdal.GDT_UInt16:
        return GDT_UInt16
    elif gdt == gdal.GDT_UInt32:
        return GDT_UInt32
    elif gdt == gdal.GDT_Unknown:
        return GDT_Unknown
    else:
        return GDT_Unknown


#!/usr/bin/env python
# coding:utf-8

"""lrdgdal Band class"""

__author__ = "Thomas Scharrenbach (thomas@scharrenbach.net)"
__copyright__ = "Copyright (C) 2014 Thomas Scharrenbach"
__license__ = "Apache License v2"
__version__ = "0.0.6"

from rdflib import Graph


class BandRDFGraph(Graph):
    """Holds a gdal band.

    This class is essentially a wrapper for a gdal band object with
    extended utility functions.
    """

    def __init__(self, band, raster, iri):
        """Create a new band holding a reference to a band.

        :param band: the actual gdal band object.
        :param iri: the iri under which this band is referenced.
        """
        super(BandRDFGraph, self).__init__(identifier=iri)
        self.__raster = raster
        self.__band = band
        self.__iri = iri

    @property
    def iri(self):
        """The band's IRI relative to it's raster's IRI."""
        return self.__iri

    @property
    def band(self):
        """The gdal band object."""
        return self.__band

    @property
    def raster(self):
        """The lrdgdal.Raster object this band belongs to."""
        return self.__raster


#!/usr/bin/env python
from setuptools import setup, find_packages

# for having nose collector work correctly.
import multiprocessing

setup(
    name='lrdgdal',
    version=open('version.txt').read(),
    author='Thomas Scharrenbach',
    author_email='thomas@scharrenbach.net',
    packages=find_packages(),
    include_package_data=True,
    url='http://scharrenbach.net',
    license='Apache License v2',
    description='Linked Raster Data for gdal for python',
    long_description=open('README.txt').read(),
    install_requires=['lxml>=3.3,',
                      'rdflib>=4.0',
                      'gdal>=1.10',
                      'numpy>=1.8'],
    setup_requires=['nose>=1.0', 'mock'],
    test_suite='nose.collector')



#!/usr/bin/env python
# coding:utf-8

"lrdgdal classes package"

__author__ = "Thomas Scharrenbach (thomas@scharrenbach.net)"
__copyright__ = "Copyright (C) 2014 Thomas Scharrenbach"
__license__ = "Apache License v2"
__version__ = "0.0.6"